'''
/** Original work Copyright 2024 damiano IU3QGD
 * https://to-be-defined
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.

 * Addition to the licence: NO company with more than 50employees can use this SW and you are NOT entitled to store this software 
 * on ANY storage that in some way attempts to monetize it (eg: github)
 * corporations and banks are too rich and too powerful, they MUST be split and ALL the money given back to the state

 
 
'''
# the idea is that this allocate the main console in a GUI
# You can have the standard console in the "main thread", this is one in a GUI
# I need a model similar to Java, for reliability, so, NO variables in module
# python has a weird idea on the scope of a module variable from other modules
# Plus, no private keyword, at all
# TO have this updated on save you need to configure the editor Save Actions
# ALSO the INITIAL VALUE MUST be valid for the given format eg: yyyy-MM-dd HH:mm 

from __future__ import annotations

import os
from threading import Thread
from tkinter import ttk, messagebox

from app_config import AppConfig, ConfigSavable
from app_csv import AppCsv
from app_gui import InputParserThread
import app_gui 
from app_version import PREPPERDOCK_version
from glob_eeprom import GlobEeprom
from glob_gui import  TkinterRoot
import glob_mlang
import intro_gui
from qconnect_gui import QconnectThread
import qpystat
from qsdtmf_gui import QsdtmfGui
from qseeprom_gui import Qseeprom_gui
from qsmsg_gui import QsmsgGui
from rabands_gui import RabandsGui
from spectrum_gui import SpectrumGui
import tkinter as tk 


# -----------------------------------------------------------------
# It has to be a thread since the main gui loop has to be in a thread
# NOTE that the remaining system setup IS in the gui run() since that is the ONLY way to make sure things are setup
# in the correct order
# do NOT set the thread to daemon, it will mess up with the input queue
class Qsk5GuiThread (Thread,ConfigSavable):

    font_size=14


    # ----------------------------------------------------------------------
    # Constructor
    def __init__(self, stat : qpystat.Qpystat):
        Thread.__init__(self,name="PrepperDock GUI Thread")

        self.stat=stat
        
        # add myself to the pieces that can be saved and restored
        self.stat.appconfig.addMyselfToSavables(self)
        
        # this first, since the GUI need to know the input queue
        self.stat.inputParser = InputParserThread(stat)

    # ----------------------------------------------------------------------
    def _println(self, msg : str ):
        self.stat.app_println(msg)

    # ----------------------------------------------------------------------
    # This is the new thread, so, we can allocate the main Gui and ALL the remaining component
    def run(self):
        print ("START Thread ", self.name)
        
        self._tkroot = TkinterRoot(self.stat, "PrepperDock", 'prepperdock', self.on_closing, False)
        
        self.menu = QuanshengMenu(self._tkroot,self.stat)
        self._tkroot.config(menu=self.menu)
        
        self._adjustTtkStyle()
        
        # the first, since the rest of the application may need it to work...        
        self.stat.glob_mlang   = glob_mlang.PrepperMlang(self.stat)

        # The magic of tkraise works ONLY with the grid layout
        # the idea is that you sticky ="nsew" and put all Frames at 0,0
        # then, you can raise them to the top with tkraise
        # HOWEVER, the panel size is the MAX dimension of all subpanels....
        self._tkroot.grid_rowconfigure(0, weight = 1)
        self._tkroot.grid_columnconfigure(0, weight = 1)

        self.stat.appmain_gui = app_gui.AppMainGui(self.stat, self._tkroot)
        self.stat.appmain_gui.grid(row = 0, column = 0, sticky ="nsew")

        self.stat.intro_gui = intro_gui.Intro_gui(self.stat,self._tkroot)
        self.stat.intro_gui.grid(row = 0, column = 0, sticky ="nsew")
        
        # ---------------------- end of gui elements that are switchable ---------------------
        
        # this will manage the debug configuration GUI
        self.stat.appcsv       = AppCsv(self.stat)
        self.stat.qconnect     = QconnectThread(self.stat)
        self.stat.qsmsg_gui    = QsmsgGui(self.stat)
        self.stat.qsdtmf_gui   = QsdtmfGui(self.stat)
        self.stat.spectrum_gui = SpectrumGui(self.stat)
        self.stat.rabands_gui  = RabandsGui(self.stat) 
        
        self.stat.globeeprom = GlobEeprom(self.stat)    # this first, the global eeprom storage
        self.stat.eeprom_gui = Qseeprom_gui(self.stat)   # then this

        # show the intro panel (could be saved)....
        self.stat.intro_gui.tkraise()
        
        # this will run half second later, to adjust windows properties and start servers
        self._tkroot.runLater(400,self._runLaterOnsettled)

        # this is basically the Swing events thread
        self._tkroot.mainloop()

        # NOTE that you MUST save window position when they are still active
        self.stat.appconfig.saveConfigToFile()    

        print ("END Thread ", self.name)
        
        # This is the most reliable way to terminate all the stuff
        os._exit(0)

    # ----------------------------------------------------------------------------
    # the part that runs on swing thread
    # NOTE that you can attach message box to different parents (and you should !!)
    def _swing_showMessageBox(self,a_message : str ):
        messagebox.showinfo("PrepperDock_message", a_message, parent=self._tkroot, type='ok')

    # ----------------------------------------------------------------------------
    # Anybody can call this one
    def showMessageBox(self, a_message : str ):
        self._tkroot.runIdle(self._swing_showMessageBox,a_message)

    # -----------------------------------------------------------------
    # this is what should be done, on this window closing        
    def on_closing(self):
        self.stat.inputParser.queuePut('quit')

    # --------------------------------------------------------------------
    # See https://tkdocs.com/tutorial/styles.html
    # AH, remember that ALL widgets styles begin with T
    # AH, remember that ttk.Entry font CANNOT be specified using a style ....
    # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/ttk-map.html
    # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/colors.html
    def _adjustTtkStyle(self):
        style = ttk.Style(self._tkroot)

        # on windows, the default theme is vista
        # and that does NOT allow to customize buttons, much
        #print(style.theme_use('clam'))

        # this changes the property for ALL widgets
        # https://www.pythontutorial.net/tkinter/ttk-style/        
        # https://wiki.tcl-lang.org/page/Changing+Widget+Colors
        style.configure(".", background='#fff9eb')
        
        style.configure("Treeview", highlightthickness=0, bd=0, font=('Calibri', self.font_size)) # Modify the font of the body
        style.configure("Treeview.Heading", font=('Calibri', self.font_size,'bold')) # Modify the font of the headings
        style.configure("TEntry", foreground = 'green', font=('Calibri', 13,'bold'), width=4) # Modify the font of the headings
        style.configure("TProgressbar", thickness=10)
        style.configure("Bold.TLabel", font=("Helvetica", 12, "bold"), foreground = '#6c2c12')
        style.configure("Header1.TLabel", font=("Helvetica", 16, "bold"), foreground = '#27abe2')
        style.configure("UMisura.TLabel", font=("Helvetica", 12, "bold"))
        style.configure("Log.TLabel", font=(12))

        style.configure("TButton", background='#faf1e6')
        
        style.configure("EEPanels.TButton", background='#d2dad0', font=('Helvetica', 12, 'bold'))
        
        #style.map('EEPanels.TButton',foreground=[('disabled', 'yellow'),('pressed', 'red'),('active', 'blue')],
        #          background=[('disabled', 'magenta'),('pressed', '!focus', 'cyan'),('active', 'green')],
        #          highlightcolor=[('focus', 'green'),('!focus', 'red')],
        #          relief=[('pressed', 'groove'),('!pressed', 'ridge')])         
         
         
        
        #style.configure("TLabel",background=def_background)
        #style.configure("TFrame", background=def_background)


    # --------------------------------------------------------------------
    # implement the config savable
    def appImportConfig(self, cfg : AppConfig ):
        self._tkroot.setGeometry(cfg.getDictValue("qsk5_gui", {}))

    # --------------------------------------------------------------------
    # implement the config savable
    def appSaveConfig(self, cfg : AppConfig ):
        cfg.setDictValue("qsk5_gui", self._tkroot.getGeometry())
        
    # ----------------------------------------------------------------
    # This is called when the GUI is up and running, basically, started on gui thread
    # with a delay of 500ms, needed to set the window status appropriately
    def _runLaterOnsettled(self):
        self.stat.appconfig.importPartConfig()
        
        # thread have to start once the GUI is up and running
        self.stat.inputParser.start()
        self.stat.qconnect.start()
        
        self._println("PrepperDock "+PREPPERDOCK_version+" started")


        
    # -------------------------------------------------------------------
    # called by another thread to shutdown the GUI
    # before shutting down, save the window position
    def endGuiConsole(self):
#        print ("endGuiConsole CALL")

        # window position must be saved before stopping GUI
        self.stat.appconfig.savePartConfig()

        # this will terminate the event loop
        # MUST be run on the gui thread to avoid concurrency issues 
        self._tkroot.runIdle(self._tkroot.destroy)

    # ------------------------------------------------------
    # Example on how you can raise different panels to top        
    def raiseMain_gui (self):
        self.stat.appmain_gui.tkraise()

            
        
# ===========================================================================        
# Want to wrap the whole menu mess in here
# This part of the gui must do nothing until all pieces are actually in place
# NOTE that the constructor creates the menu structure

class QuanshengMenu(tk.Menu):

    # ------------------------------------------------------
    def __init__(self, guiroot, stat : qpystat.Qpystat):
        tk.Menu.__init__(self,guiroot)
        
        self.stat=stat;
        
        self.add_command(label="Home", command=self._showIntro)
        self.add_command(label="Advanced", command=self._showConsole)
        #self.add_command(label="Languages", command=self._selectResources)
        
        # if you want to create a hierarchy. create a Menu with this parent
        #menutools = tk.Menu(self,tearoff=0)
        
        #menutools.add_command(label="Show Console", command=self._showConsole)
        #menutools.add_command(label="Show Intro", command=self._showIntro)
    
        #self.add_cascade(label="File", menu=menutools)

    # ------------------------------------------------------
    # should allow a user to select the topdir where resources are stored
    # normally in user home directory
    def _selectResources(self):
        self.stat.glob_mlang.GUI_show_window()
    
    # ------------------------------------------------------
    def _showConsole(self):
        self.stat.app_println("show Console")
        self.stat.qsk5_gui.raiseMain_gui()

    # ------------------------------------------------------
    def _showIntro(self):
        self.stat.app_println("show Intro")
        self.stat.intro_gui.tkraise()


