'''
/** 
 * Original work Copyright 2024 damiano IU3QGD
 * https://to-be-defined
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.

 * Addition to the licence: NO company with more than 50employees can use this SW and you are NOT entitled to store this software 
 * on ANY storage that in some way attempts to monetize it (eg: github)
 * corporations and banks are too rich and too powerful, they MUST be split and ALL the money given back to the state

Squelch, this is mostly to handle squelch level 0
But could be extended easly to handle all squelch 

  
''' 

from __future__ import annotations

from tkinter import ttk, IntVar, messagebox
import typing

from app_csv import CsvTable, Csv_Writer
from glob_eeprom import GlobEeprom, EEblock
from glob_fun import getResourcesPhotoimage
from glob_gui import SolarTableColdef, EE_events_listener, GridPanel, TV_Entry, SolarTable, table_row_selected_background_color, PackPanelLeft
import qpystat


# The # type: ignore is for mypy to stop complaining that it cannot find the correct tyep
qsquelchcol_idx="qsquelchcol_idx"
qsquelchcol_band="qsquelchcol_band"
qsquelchcol_level="qsquelchcol_level"
qsquelchcol_rssi_gt_open="qsquelchcol_rssi_gt_open"
qsquelchcol_rssi_lt_close="qsquelchcol_rssi_lt_close"  
qsquelchcol_noise_lt_open="qsquelchcol_noise_lt_open"  
qsquelchcol_noise_gt_close="qsquelchcol_noise_gt_close"         
qsquelchcol_glitch_lt_open="qsquelchcol_glitch_lt_open"
qsquelchcol_glitch_gt_close="qsquelchcol_glitch_gt_close"
    
# -------------------------------------------------------
# Generic way to define a map from a column name to some properties        
qsquelchcol_map = { 
    qsquelchcol_idx            : SolarTableColdef(qsquelchcol_idx        ,'Nr.',True, 40), 
    qsquelchcol_band           : SolarTableColdef(qsquelchcol_band       ,"Band",False,50),
    qsquelchcol_level          : SolarTableColdef(qsquelchcol_level      ,"Level",False, 50),
    qsquelchcol_rssi_gt_open   : SolarTableColdef(qsquelchcol_rssi_gt_open   ,"RSSI > O",False, 80),
    qsquelchcol_rssi_lt_close  : SolarTableColdef(qsquelchcol_rssi_lt_close  ,"RSSI < C", False, 80),
    qsquelchcol_noise_lt_open  : SolarTableColdef(qsquelchcol_noise_lt_open  ,"Noise < O",False,90),
    qsquelchcol_noise_gt_close : SolarTableColdef(qsquelchcol_noise_gt_close ,"Noise > C",False,90),
    qsquelchcol_glitch_lt_open : SolarTableColdef(qsquelchcol_glitch_lt_open ,"Glitch < O",False,90),
    qsquelchcol_glitch_gt_close: SolarTableColdef(qsquelchcol_glitch_gt_close,"Glitch > C",False,90),
    }


# =====================================================================
# the table with the list of FM channels
# It should take the data from the global EEPROM repository
# It will be signalled when new EEPROM data is received
class Qsquelch_gui (EE_events_listener):

    # ----------------------------------------------------------------
    def __init__(self, stat : qpystat.Qpystat, parent_frame : ttk.Frame ):
        self.stat = stat

        self.work_panel = ttk.Frame(parent_frame)
        
        self._qsquelch_list : Qsquelch_list = Qsquelch_list()
        self.stat.appcsv.add_to_tablesList(self._qsquelch_list)
        
        self.c_panel=self._newCenterPanel(self.work_panel)
        
        self.c_panel.pack(fill='both', expand=True)


    # ------------------------------------------------------------------------
    def tkraise(self):
        self.work_panel.tkraise()
        
    # ------------------------------------------------------------------------
    def _println(self, msg : str ):
        self.stat.app_println(msg)
        
    # -------------------------------------------------------------------------------
    # On center panel there is the table and scroll bar
    # NOTE: The layout of this is described by caller
    # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/ttk-Treeview.html
    def _newCenterPanel(self, parent):
        apanel = ttk.Frame( parent) #, borderwidth=3, relief='ridge')
        
        #apanel.grid_rowconfigure(0, weight = 1)
        #apanel.grid_columnconfigure(0, weight = 1)
        
        self._table_view = SolarTable(apanel, qsquelchcol_map );
        self._table_view.getComponent().pack(fill='both', expand=True)

        tk_tbl : ttk.Treeview = self._table_view.getTable()
        
        tk_tbl.tag_configure('Updated',  background=table_row_selected_background_color)
        tk_tbl.bind('<<TreeviewSelect>>', self._jtable_row_selected)
        
        # --------------------------------------------------- now, the editing panel
        self._editingPanel = Qsquelch_edit_gui(self ,apanel)
        self._editingPanel.getComponent().pack(fill='x')
        
        return apanel

    # -------------------------------------------------------------------
    # Called when a row is selected in the tree
    # note that you NEED selectmode='browse' to have a single selection
    # IF multiple rows can be selected with option selectmode='extended' you get a list of selected items that can be not in sequence
    # you can use SHIFT or CTRRL + click to add or remove selected elements
    def _jtable_row_selected(self, sel_coord):
        
        a_sel = self._table_view.getSelected_iid()
        
        if not a_sel:
            return 
        
        #self._println(str(a_sel))
        
        row_idx : int = int(a_sel[0])
        
        sel_item = self._qsquelch_list.get_row_using_idx(row_idx)

        self._editingPanel.showThisRow(sel_item)
        
    # ------------------------------------------------------------------------
    def showInfo(self, a_message : str ):
        messagebox.showinfo("Channels", a_message, parent=self.work_panel, type='ok')    
        
    # ------------------------------------------------------------------------
    # the RSSI table is quite messy, let me try to use the radio logic...
    def updateGuiFromEeprom(self):
        self._qsquelch_list.updateListFromEeprom(self.stat.globeeprom)
        
        self.GUI_refresh()



    # -----------------------------------------------------------------
    # When this is called, it means the user pressed the write button
    def update_EEblock_from_GUI(self):

        self._println("qsquelch.uupdate_EEblock_from_GUI CALL")

        for row in self._qsquelch_list.qsquelch_list:
            row.update_EEblock_from_GUI(self.stat.globeeprom)
            
        self.GUI_refresh()


    # -------------------------------------------------------------------
    # At any time it should be possible to filter the entries and show them
    # Note that this may be called at the beginning , when there is nothing in the list...
    def GUI_refresh(self ):    
        
        for row in self._qsquelch_list.qsquelch_list:
            e_row : Qsquelch_row = row
            
            self._jtable_ensureVisible(e_row)

    # ------------------------------------------------------------------------
    # Utility, to make code cleaner
    # NOTE that part of the magic works because the channel number is a unique key
    # This MUST be run from swing thread
    def _jtable_ensureVisible(self, a_row : Qsquelch_row ):

        if self._table_view.row_exist(a_row.row_idx):
            
            self._jtable_refresh_using_row(a_row)
                
            #self._table_view.see_row(a_row.qsch_idx_po)
        else:
            v_values=a_row.getValuesForTreeview()
            self._table_view.row_append(a_row.row_idx,v_values)



    # ------------------------------------------------------------------------------
    # this should update the swing table content for the given row
    def _jtable_refresh_using_row(self, a_row : Qsquelch_row ):

        if not a_row:
            return

        #self._table_view.focus_row(a_row.qsch_idx_po)
        
        updated_str=''
        
        if a_row.isUserModified:
            updated_str='Updated'
            
        self._table_view.set_row(a_row.row_idx,a_row.getValuesForTreeview(),updated_str)   


# ==============================================================================
# Q squelch row is made of the following components
# note that there is a value for the band and also for the level
# also, there is something on a delay for squelch... on reg_4E
class Qsquelch_row():

    csv_heading=['sq_idx',
                 'sq_band',
                 'sq_lvl',
                 'sq_rssi_gt_open',
                 'sq_rssi_lt_close',
                 'sq_noise_lt_open',
                 'sq_noise_gt_close',
                 'sq_glitch_lt_open',
                 'sq_glitch_gt_close']

    # --------------------------------------------------------------------------
    def __init__(self, r_idx : int, vhf_uhf : str, squelch_lvl : int ):

        self.row_idx : int = r_idx       # row index, to pick up values from the table click
        self.is_vhf_uhf : str = vhf_uhf
        self.squelch_level : int = squelch_lvl
        
        self.rssi_gt_open : int       # a rssi measured above this will open
        self.rssi_lt_close : int      # a rssi below this will close
        
        self.noise_lt_open : int     
        self.noise_gt_close : int    
        
        self.glitch_lt_open : int     
        self.glitch_gt_close : int    

        self.isUserModified=False


    # -----------------------------------------------------------------
    # return the list of values to show into the table, the GUI table
    # NOTE that the order has to be consistent with the GUI definition
    def getValuesForTreeview(self):
        
        risul = (
            self.row_idx,
            self.is_vhf_uhf,
            self.squelch_level,
            self.rssi_gt_open,
            self.rssi_lt_close,
            self.noise_lt_open,
            self.noise_gt_close,
            self.glitch_lt_open,
            self.glitch_gt_close
            )
            
        return risul         

        
    # -----------------------------------------------------------------
    # must return the values as a list of strings in the same order as the header
    def csv_get_values(self) -> typing.List[str]:
        
        risul : typing.List[str] = []
        
        risul.append(str(self.row_idx))
        risul.append(str(self.is_vhf_uhf))
        risul.append(str(self.squelch_level))
        
        risul.append(str(self.rssi_gt_open))
        risul.append(str(self.rssi_lt_close))
        
        risul.append(str(self.noise_lt_open))
        risul.append(str(self.noise_gt_close))

        risul.append(str(self.glitch_lt_open))
        risul.append(str(self.glitch_gt_close))
        
        return risul
        
        
    # ------------------------------------------------------------------
    # this assumes tha values are in the SAME order and same format as the export !
    # note that s_split is an iterator for strings and the FIRST item (the row index) has already been taken out
    def csv_parse_row(self, r_iter ):
        if not r_iter:
            return 
        
        if self.is_vhf_uhf != next(r_iter):
            return 
        
        if self.squelch_level != next(r_iter):
            return 
        
        self.rssi_gt_open = int(next(r_iter))
        self.rssi_lt_close = int(next(r_iter))
        
        self.noise_lt_open = int(next(r_iter))
        self.noise_gt_close = int(next(r_iter))
        
        self.glitch_lt_open = int(next(r_iter))
        self.glitch_gt_close = int(next(r_iter))

        # a paste from clipboard is equivalent of a user modified
        self.isUserModified=True
        
    # -----------------------------------------------------------------
    def _getBlockByteValue(self, g_eeprom : GlobEeprom, base_address : int, param_index : int, byte_index : int ):
        
        v_address = base_address + param_index * 16
        
        a_block : EEblock = g_eeprom.get_block_from_address(v_address)
        
        return a_block.get_byte(byte_index)
        
    # =-----------------------------------------------------------------
    def _getSquelchBase(self) -> int:
        if self.is_vhf_uhf == 'UHF':
            return 0x1E00
        else:
            return 0x1E60

    # =-----------------------------------------------------------------
    # values are spread in different bloks with different start address        
    def updateRowFromEeprom(self, g_eeprom : GlobEeprom ):
        
        base_addres = self._getSquelchBase()
        
        self.rssi_gt_open  = self._getBlockByteValue(g_eeprom, base_addres  ,0 ,self.squelch_level)
        self.rssi_lt_close = self._getBlockByteValue(g_eeprom, base_addres  ,1 ,self.squelch_level)

        self.noise_lt_open  = self._getBlockByteValue(g_eeprom, base_addres ,2 ,self.squelch_level)
        self.noise_gt_close = self._getBlockByteValue(g_eeprom, base_addres ,3 ,self.squelch_level)
        
        self.glitch_lt_open  = self._getBlockByteValue(g_eeprom, base_addres ,4 ,self.squelch_level)
        self.glitch_gt_close = self._getBlockByteValue(g_eeprom, base_addres ,5 ,self.squelch_level)
        
        self.isUserModified=False

    # -----------------------------------------------------------------
    def _setBlockByteValue(self, g_eeprom : GlobEeprom, base_address : int, param_index : int, byte_index : int, b_value : int ):
        
        v_address = base_address + param_index * 16
        
        a_block : EEblock = g_eeprom.get_block_from_address(v_address)
        
        return a_block.set_user_byte(byte_index, b_value)

    # -----------------------------------------------------------------
    # Note that what can be changed are just the values, 
    # update ONLY if the data has changed
    def update_EEblock_from_GUI(self, g_eeprom : GlobEeprom ):
        
        if not self.isUserModified:
            return

        # once written to EEPROM block I can consider it clear
        self.isUserModified=False

        base_addres = self._getSquelchBase()

        self._setBlockByteValue(g_eeprom, base_addres  ,0 ,self.squelch_level, self.rssi_gt_open)
        self._setBlockByteValue(g_eeprom, base_addres  ,1 ,self.squelch_level, self.rssi_lt_close)
        
        self._setBlockByteValue(g_eeprom, base_addres  ,2 ,self.squelch_level, self.noise_lt_open)
        self._setBlockByteValue(g_eeprom, base_addres  ,3 ,self.squelch_level, self.noise_gt_close)
        
        self._setBlockByteValue(g_eeprom, base_addres  ,4 ,self.squelch_level, self.glitch_lt_open)
        self._setBlockByteValue(g_eeprom, base_addres  ,5 ,self.squelch_level, self.glitch_gt_close)
        
        


        
# ========================================================================
# Let me try to have some decent OO wrapping in this shitty language
# The raw background data is always allocated
class Qsquelch_list(CsvTable):

    # -----------------------------------------------------------------
    # I need to preallocate list and never deallocate them
    # NOTE that squelch goes from zero to 9, inclusive
    def __init__(self):
        
        self.qsquelch_list : typing.List[Qsquelch_row] = []
        
        r_idx = 0
        
        for s_band in ['VHF','UHF']:
            for s_level in range(10):
                self.qsquelch_list.append(Qsquelch_row(r_idx,s_band,s_level))
                r_idx += 1
        
    # -----------------------------------------------------------------------
    # if idx is out of range, return None
    def get_row_using_idx(self, idx : int ) -> Qsquelch_row:
        return self.qsquelch_list[idx]
        
    # -----------------------------------------------------------------------
    def updateListFromEeprom(self, g_eeprom : GlobEeprom ):
        for row in self.qsquelch_list:
            row.updateRowFromEeprom(g_eeprom)
        
    # ----------------------------------------------------------
    # override CSV
    def csv_tableName(self) -> str:
        return 'radio_squelch'
    
    # ----------------------------------------------------------
    # override CSV
    def csv_write_header(self, a_writer : Csv_Writer):
        a_writer.writerow(Qsquelch_row.csv_heading)

    # ---------------------------------------------------------
    # in a different thread than swing, write the table to this csv writer
    def csv_write_table(self, a_writer : Csv_Writer):
        
        for row in self.qsquelch_list:
            v_list = row.csv_get_values()
            a_writer.writerow(v_list)

    # -------------------------------------------------------------
    # the difference is that clipboard will insert from a given point
    # for CSV, instead, I am given the row index            
    def csv_parse_row(self, row_values : typing.List[str] ):
        r_iter = iter(row_values)
        
        r_idx = int(next(r_iter))
        
        a_row : Qsquelch_row = self.qsquelch_list[r_idx]
        
        a_row.csv_parse_row(r_iter)

        
        
# ==============================================================================
# it is kind of easier to have a separate class            
class Qsquelch_edit_gui():


    # ------------------------------------------------------------
    # I make this a frame for ease of use, it should a bit hidden...    
    def __init__(self, parent_class : Qsquelch_gui , parent_panel : ttk.Frame ):
        
        self.parent_class = parent_class

        # need an empty row, by default
        self.sel_item : Qsquelch_row = Qsquelch_row(0,'',0)

        self._work_panel = self._newWorkPanel(parent_panel)

    # ------------------------------------------------------------
    def getComponent(self) -> ttk.Frame:
        return self._work_panel

    def _newWorkPanel(self, parent_panel : ttk.Frame ) -> ttk.Frame:
        apanel = PackPanelLeft(parent_panel)
        
        apanel.addItem(self._newGridPanel(apanel))

        self._a_i1=getResourcesPhotoimage('save-64x64.png')
        apanel.addItem(ttk.Button(apanel, image=self._a_i1, command=self._clicSaveEditSquelch ))
        
        return apanel

    # ------------------------------------------------------------
    def _println(self, msg : str ):
        self.parent_class._println(msg)

    # -------------------------------------------------------------------------------
    def _newGridPanel(self, parent_panel) -> GridPanel:
        gridpanel = GridPanel(parent_panel) # , borderwidth=1, relief='solid', padding=3)

        self._edit_rssi_gt_open = IntVar(gridpanel)
        self._edit_rssi_lt_close = IntVar(gridpanel)

        self._edit_noise_lt_open = IntVar(gridpanel)
        self._edit_noise_gt_close = IntVar(gridpanel)

        self._edit_glitch_lt_open = IntVar(gridpanel)
        self._edit_glitch_gt_close = IntVar(gridpanel)


        # ------------------------------------------------------------
        gridpanel.nextRow()
        gridpanel.addItem(ttk.Label(gridpanel,text="RSSI > open"))
        gridpanel.addItem(TV_Entry(gridpanel, self._edit_rssi_gt_open,width=4))

        gridpanel.addItem(ttk.Label(gridpanel,text="RSSI < close"))
        gridpanel.addItem(TV_Entry(gridpanel, self._edit_rssi_lt_close,width=4))

        gridpanel.addItem(ttk.Label(gridpanel,text="Max 255, 0.5dB/step"))

        gridpanel.nextRow()

        gridpanel.addItem(ttk.Label(gridpanel,text="Noise < open"))
        gridpanel.addItem(TV_Entry(gridpanel, self._edit_noise_lt_open,width=4))

        gridpanel.addItem(ttk.Label(gridpanel,text="Noise > close"))
        gridpanel.addItem(TV_Entry(gridpanel, self._edit_noise_gt_close,width=4))

        gridpanel.addItem(ttk.Label(gridpanel,text="Max 127"))

        gridpanel.nextRow()

        gridpanel.addItem(ttk.Label(gridpanel,text="Glitch < open"))
        gridpanel.addItem(TV_Entry(gridpanel, self._edit_glitch_lt_open,width=4))

        gridpanel.addItem(ttk.Label(gridpanel,text="Glitch > close"))
        gridpanel.addItem(TV_Entry(gridpanel, self._edit_glitch_gt_close,width=4))

        gridpanel.addItem(ttk.Label(gridpanel,text="Max 255"))

        return gridpanel

        

    # -------------------------------------------------------------------------------
    # The user click SAVE in the edit panel
    # this should map the values back to the selected row
    def _clicSaveEditSquelch(self):

        if not self.sel_item:
            self.parent_class.showInfo("_clicSaveEditSquelch: Please select a row")

        self.sel_item.rssi_gt_open = self._edit_rssi_gt_open.get()
        self.sel_item.rssi_lt_close = self._edit_rssi_lt_close.get()

        self.sel_item.noise_lt_open = self._edit_noise_lt_open.get()
        self.sel_item.noise_gt_close = self._edit_noise_gt_close.get()

        self.sel_item.glitch_lt_open = self._edit_glitch_lt_open.get()
        self.sel_item.glitch_gt_close = self._edit_glitch_gt_close.get()
        
        self.sel_item.isUserModified=True

        self.parent_class._jtable_refresh_using_row(self.sel_item)

    # -------------------------------------------------------------------------------
    # Parent request to show this row
    def showThisRow(self, sel_item : Qsquelch_row ):

        if not sel_item:
            self.parent_class.showInfo("showThisRow: Please select a row")
            return;
        
        self.sel_item=sel_item  # will be used in the save
        
        self._edit_rssi_gt_open.set(sel_item.rssi_gt_open)
        self._edit_rssi_lt_close.set(sel_item.rssi_lt_close)
        
        self._edit_noise_lt_open.set(sel_item.noise_lt_open)
        self._edit_noise_gt_close.set(sel_item.noise_gt_close)

        self._edit_glitch_lt_open.set(sel_item.glitch_lt_open)
        self._edit_glitch_gt_close.set(sel_item.glitch_gt_close)


                

        
        
        


